/* This is an example to programme userdata. This library implements a single-linked list.

   See also: http://www.martinbroadhurst.com/data-structures.html */

#define slist_c
#define LUA_LIB

#include <stdlib.h>
#include <string.h>

#include "agena.h"
#include "agnxlib.h"
#include "agenalib.h"
#include "lgc.h"
#include "lobject.h"
#include "lstate.h"
#include "lapi.h"
#include "lmem.h"

#define checkslist(L, n) (Slist *)luaL_checkudata(L, n, "slist")
#define checkslistnode(L, n) (Slnode *)luaL_checkudata(L, n, "slistnode")


#if !(defined(LUA_DOS) || defined(__OS2__) || defined(LUA_ANSI))
#define AGENA_SLISTLIBNAME "slist"
LUALIB_API int (luaopen_slist) (lua_State *L);
#endif


struct Slnode {
  int data;  /* actually a reference created by luaL_ref into the registry which stores the original Lua value */
  struct Slnode *next;
};

typedef struct Slnode Slnode;

typedef struct {
  Slnode *head;
  Slnode *tail;
  size_t size;
} Slist;


Slnode *sl_createnode (lua_State *L, int idx) {
  Slnode *node = malloc(sizeof(Slnode));
  if (node) {
    lua_pushvalue(L, idx);
    node->data = luaL_ref(L, LUA_REGISTRYINDEX);
    node->next = NULL;
  }  
  return node;
}

void sl_empty (Slist *list) {
  Slnode *node, *temp;
  node = list->head;
  while (node != NULL) {
    temp = node->next;
    xfree(node);
    node = temp;
  }
}

void sl_delete (Slist *list) {
  if (list) {
    sl_empty(list);
    xfree(list);
  }
}

void sl_append (Slist *list, Slnode *node) {
  if (list->head == NULL) {  /* at the very first node */
    list->head = node;
    list->tail = node;
  } else {
    list->tail->next = node;
    list->tail = node;
  }
  list->size++;
}

void sl_prepend (Slist *list, Slnode *node) {
  if (list->tail == NULL) {  /* at the very first node */
    list->head = node;
    list->tail = node;
  } else {
    node->next = list->head;
    list->head = node;
  }
  list->size++;
}

void sl_removehead (Slist *list) {
  if (list->head) {
    Slnode *temp = list->head;
    list->head = list->head->next;
    if (list->head == NULL) {  /* List is now empty */
      list->tail = NULL;
    }
    xfree(temp);
    list->size--;
    if (list->size == 1) {
      list->tail = list->head;
    }
  }
}

/* static void sl_removetail (Slist *list) {
  if (list->tail) {
    MBsnode *current, *previous = NULL;
    current = list->head;
    while (current->next) {
      previous = current;
      current = current->next;
    }
    xfree(current->data);
    xfree(current);
    if (previous) {
      previous->next = NULL;
      list->tail = previous;
    } else { // List is now empty
      list->head = NULL;
      list->tail = NULL;
    }
    list->size--;
    if (list->size == 1) {
      list->head = list->tail;
    }
  }
} */


static int slist_list (lua_State *L) {  /* create a new userdata slist  */
  size_t i, nargs;
  Slist *list;
  Slnode *node;
  nargs = lua_gettop(L);
  luaL_checkstack(L, nargs, "too many arguments");
  list = (Slist *)lua_newuserdata(L, sizeof(Slist));
  if (list) {
    list->head = NULL;
    list->tail = NULL;
    list->size = 0;
  } else
    luaL_error(L, "Error in " LUA_QS ": memory allocation error.", "slist.list");
  luaL_getmetatable(L, "slist");  /* assign the metatable defined below to the userdata */
  lua_setmetatable(L, -2);
  for (i=1; i <= nargs; i++) {
    lua_lock(L);
    node = sl_createnode(L, i);
    if (node == NULL)
      luaL_error(L, "Error in " LUA_QS ": memory allocation error.", "slist.list");
    sl_append(list, node);
    lua_unlock(L);
  }
  return 1;
}


static int slist_node (lua_State *L, Slnode *node) {  /* create a new userdata slist node */
  Slnode *newnode;
  lua_lock(L);
  if (node == NULL) {
    lua_pushnil(L);
    return 1;
  }
  newnode = (Slnode *)lua_newuserdata(L, sizeof(Slnode));
  if (newnode) {
    newnode->data = node->data;
    newnode->next = node->next;
  } else
    luaL_error(L, "Error in " LUA_QS ": memory allocation error.", "slist.node");
  lua_unlock(L);
  luaL_getmetatable(L, "slistnode");  /* assign the metatable defined below to the userdata */
  lua_setmetatable(L, -2);
  return 1;  /* pushed node userdata on stack */
}


static int slist_append (lua_State *L) {
  Slist *list;
  Slnode *node;
  size_t nargs, i;
  nargs = lua_gettop(L);
  luaL_checkstack(L, nargs, "too many arguments");
  if (lua_gettop(L) < 2)
    luaL_error(L, "Error in " LUA_QS ": need at least an slist and a value.", "slist.append");
  list = (Slist *)checkslist(L, 1);
  for (i=2; i <= nargs; i++) {
    lua_lock(L);
    node = sl_createnode(L, i);
    if (node == NULL)
      luaL_error(L, "Error in " LUA_QS ": memory allocation error.", "slist.append");
    sl_append(list, node);
    lua_unlock(L);
  }
  return 0;
}


static int slist_prepend (lua_State *L) {
  Slist *list;
  Slnode *node;
  size_t i, nargs;
  nargs = lua_gettop(L);
  luaL_checkstack(L, nargs, "too many arguments");
  if (lua_gettop(L) < 2)
    luaL_error(L, "Error in " LUA_QS ": need at least an slist and a value.", "slist.prepend");
  list = (Slist *)checkslist(L, 1);
  for (i=2; i <= nargs; i++) {
    lua_lock(L);
    node = sl_createnode(L, i);
    if (node == NULL)
      luaL_error(L, "Error in " LUA_QS ": memory allocation error.", "slist.prepend");
    sl_prepend(list, node);
    lua_unlock(L);
  }
  return 0;
}


static int slist_put (lua_State *L) {
  Slist *list;
  Slnode *newnode;
  int idx;
  if (lua_gettop(L) != 3)
    luaL_error(L, "Error in " LUA_QS ": need an slist, an index, and any value.", "slist.put");
  lua_lock(L);
  list = (Slist *)checkslist(L, 1);
  idx = luaL_checkint(L, 2);
  if (idx < 1 || idx > list->size + 1)
    luaL_error(L, "Error in " LUA_QS ": index %d is out of range.", "slist.put", idx);
  newnode = sl_createnode(L, 3);
  if (newnode == NULL)
    luaL_error(L, "Error in " LUA_QS ": memory allocation error.", "slist.put");
  if (list->size + 1 == idx) {
    sl_append(list, newnode);
  } else {
    size_t i;
    Slnode *node, *previous;
    previous = NULL;
    node = list->head;
    for (i=1; i < idx; i++) {
      previous = node;
      node = node->next;
    }
    if (node == list->head) {
      if (list->tail == NULL) {  /* Adding the first node */
        list->head = newnode;
        list->tail = newnode;
      } else {
		newnode->next = list->head;
		list->head = newnode;
	  }
    } else {
      newnode->next = node;
      previous->next = newnode;
    }
    list->size++;
  }
  return 0;
}


static void sl_remove (Slist *list, Slnode *node, Slnode *previous) {
  if (node == list->head) {
    sl_removehead(list);
  } else {
    previous->next = node->next;
    list->size--;
    if (list->size == 1) {
	  list->tail = list->head;
    } else if (node == list->tail) {
	  list->tail = previous;
    }
    xfree(node);
  }
}

static int slist_purge (lua_State *L) {
  Slist *list;
  Slnode *previous, *node;
  int idx;
  size_t i;
  if (lua_gettop(L) != 2)
    luaL_error(L, "Error in " LUA_QS ": need an slist, an index, and any value.", "slist.purge");
  lua_lock(L);
  list = (Slist *)checkslist(L, 1);
  idx = luaL_checkint(L, 2);
  if (idx < 1 || idx > list->size)
    luaL_error(L, "Error in " LUA_QS ": index %d is out of range.", "slist.purge", idx);
  previous = NULL;
  node = list->head;
  for (i=1; i < idx; i++) {
    previous = node;
    node = node->next;
  }
  sl_remove(list, node, previous);
  return 0;
}


int iterate (lua_State *L) {
  Slnode *node;
  node = (Slnode *)lua_touserdata(L, lua_upvalueindex(1));
  if (node == NULL) {
    lua_pushnil(L);
  } else {
    lua_rawgeti(L, LUA_REGISTRYINDEX, node->data);
    slist_node(L, node->next);
    lua_replace(L, lua_upvalueindex(1));  /* update upvalue */
  }
  return 1;
}


static int slist_iterate (lua_State *L) {
  Slist *list;
  Slnode *node;
  int i, pos;
  list = checkslist(L, 1);
  pos = luaL_optint(L, 2, 1);
  if (pos < 1)
    luaL_error(L, "Error in " LUA_QS ": start index must be positive.", "slist.iterate");
  node = list->head;
  for (i=1; node != NULL && i < pos; i++) {
    node = node->next;
  }
  slist_node(L, node);    /* push node userdata */
  lua_pushcclosure(L, &iterate, 1);
  return 1;
}


static int slist_getsize (lua_State *L) {  /* return the number of allocated slots in the userdata numeric array */
  Slist *list = checkslist(L, 1);
  lua_pushinteger(L, list->size);
  return 1;
}


static int slist_getitem (lua_State *L) {  /* get the value at the given index position */
  size_t i, index;
  Slnode *node;
  Slist *list = checkslist(L, 1);
  index = luaL_checkint(L, 2);
  if (index < 1 || index > list->size) {
    lua_pushnil(L);
    return 1;
  }
  node = list->head;
  for (i=1; i < index; i++) {
    node = node->next;
  }
  lua_rawgeti(L, LUA_REGISTRYINDEX, node->data);
  return 1;
}


static int slist_in (lua_State *L) {
  int r, nargs;
  Slnode *node;
  nargs = lua_gettop(L);
  if (nargs != 2) {
    luaL_error(L, "Error in " LUA_QS ": need two arguments.", "slist.__in");
  }
  r = 0;
  Slist *list = checkslist(L, 2);
  for (node = list->head; node != NULL; node = node->next) {
    //setobj2s(L, L->top++, node->data);
    lua_rawgeti(L, LUA_REGISTRYINDEX, node->data);
    if ((r = lua_equal(L, 1, -1)) == 1) {
      //setnilvalue(L->top - 1); L->top--;
      //setnilvalue(L->top); L->top--;
      agn_poptop(L);
      break;
    }
    //L->top--;
    agn_poptop(L);
  }
  lua_pushboolean(L, r);
  return 1;
}


static int slist_setitem (lua_State *L) {  /* set a value at the given index position */
  size_t i, idx;
  Slnode *node, *newnode;
  if (lua_gettop(L) != 3)
    luaL_error(L, "Error in " LUA_QS ": need an slist, an index, and a value.", "slist.__writeindex");
  Slist *list = checkslist(L, 1);
  idx = luaL_checkint(L, 2);
  if (idx < 1 || idx > list->size + 1) {
    lua_pushnil(L);
    return 0;
  }
  newnode = sl_createnode(L, 3);
  if (newnode == NULL)
    luaL_error(L, "Error in " LUA_QS ": memory allocation error.", "slist.__writeindex");
  if (list->size + 1 == idx) {
    sl_append(list, newnode);
  } else {  /* replace existing item */
    node = list->head;
    for (i=1; i < idx; i++) {
      node = node->next;
    }
    node->data = newnode->data;
  }
  return 0;
}


static int slist_tostring (lua_State *L) {  /* at the console, the array is formatted as follows: */
  (void)checkslist(L, 1);
  lua_pushfstring(L, "slist(%p)", lua_topointer(L, 1));
  return 1;
}


static int slist_gc (lua_State *L) {  /* please do not forget to garbage collect deleted userdata */
  (void)L;
//  Slist *list = checkslist(L, 1);
//  if (list->head != NULL) sl_delete(list);
  lua_pushnil(L);
  lua_setmetatable(L, 1);
  return 0;
}


static int slist_nodegc (lua_State *L) {  /* please do not forget to garbage collect deleted userdata */
  (void)L;
  return 0;
}


static const struct luaL_Reg slist_slistlib [] = {  /* metamethods for linked lists `n' */
  {"__index", slist_getitem},       /* n[p], with p the index, counting from 1 */
  {"__writeindex", slist_setitem},  /* n[p] := value, with p the index, counting from 1 */
  {"__size", slist_getsize},        /* retrieve the number of entries in `n' */
  {"__tostring", slist_tostring},   /* for output at the console, e.g. print(n) */
  {"__in", slist_in},               /* `in` operator for linked lists */
  {"__gc", slist_gc},               /* please do not forget garbage collection */
  {NULL, NULL}
};


static const struct luaL_Reg slist_slistnodelib [] = {  /* metamethods for numeric userdata `n` */
  {"__gc", slist_nodegc},              /* please do not forget garbage collection */
  {NULL, NULL}
};


static const luaL_Reg slistlib[] = {
  {"append", slist_append},
  {"iterate", slist_iterate},
  {"list", slist_list},
  {"prepend", slist_prepend},
  {"purge", slist_purge},
  {"put", slist_put},
  {NULL, NULL}
};


static void createmetatable (lua_State *L) {
  lua_createtable(L, 0, 1);  /* create metatable for strings */
  lua_pushliteral(L, "");    /* dummy string */
  lua_pushvalue(L, -2);
  lua_setmetatable(L, -2);   /* set string metatable */
  agn_poptop(L);             /* pop dummy string */
  lua_pushvalue(L, -2);      /* string library... */
  lua_setfield(L, -2, "__index");  /* ...is the __index metamethod */
  agn_poptop(L);             /* pop metatable */
}


/*
** Open slist library
*/
LUALIB_API int luaopen_slist (lua_State *L) {
  /* metamethods for linked lists */
  luaL_newmetatable(L, "slist");
  luaL_register(L, NULL, slist_slistlib);
  /* metamethod for the nodes (for `slist.iterate`) */
  luaL_newmetatable(L, "slistnode");
  luaL_register(L, NULL, slist_slistnodelib);
  luaL_register(L, AGENA_SLISTLIBNAME, slistlib);
  createmetatable(L);
  return 1;
}


